"""Tests for filename parsing."""

import pytest

from comicfn2dict import ComicFilenameSerializer
from tests.comic_filenames import SERIALIZE_FNS


@pytest.mark.parametrize("item", SERIALIZE_FNS.items())
def test_serialize_dict(item):
    """Test metadata serialization."""
    test_fn, md = item
    fn = ComicFilenameSerializer(md).serialize()
    assert test_fn == fn



"""Test filenames with human parsed correct results."""

from types import MappingProxyType

TEST_COMIC_FIELDS = {
    "series": "Long Series Name",
    "issue": "001",
    "year": "2000",
    "original_format": "TPB",
    "scan_info": "Releaser",
    "title": "Title",
    "ext": "cbz",
}
TEST_COMIC_FIELDS_VOL = {
    "series": "Long Series Name",
    "volume": "1",
    "issue": "001",
    "year": "2000",
    "title": "Title",
    "original_format": "TPB",
    "scan_info": "Releaser",
    "ext": "cbz",
}
TEST_COMIC_VOL_ONLY = {
    "series": "Long Series Name",
    "volume": "1",
    "issue": "1",
    "title": "Title",
    "original_format": "TPB",
    "year": "2000",
    "scan_info": "Releaser & Releaser-Releaser",
    "ext": "cbr",
}

# Tests for 0.1.0
FNS = {
    "Night of 1000 Wolves 001 (2013).cbz": {
        "series": "Night of 1000 Wolves",
        "issue": "001",
        "year": "2013",
        "ext": "cbz",
    },
    "19687 Sandman 53.cbz": {
        "series": "19687 Sandman",
        "issue": "53",
        "ext": "cbz",
    },
    "33475 OMAC v3 2.cbr": {
        "series": "33475 OMAC",
        "volume": "3",
        "issue": "2",
        "ext": "cbr",
    },
    "Long Series Name 001 (2000) Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name #001 (2000) Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name (2000) 001 Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name (2000) #001 Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Ultimate Craziness (2019) (Digital) (Friends-of-Bill).cbr": {
        "series": "Ultimate Craziness",
        "year": "2019",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Friends-of-Bill",
    },
    "Jimmy Stocks Love Chain (2005) (digital) (The Magicians-Empire).cbr": {
        "series": "Jimmy Stocks Love Chain",
        "year": "2005",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "The Magicians-Empire",
    },
    "Arkenstone Vol. 01 - The Smell of Burnt Toast (2020) (digital) (My-brother).cbr": {
        "series": "Arkenstone",
        "volume": "01",
        "issue": "01",
        "year": "2020",
        "ext": "cbr",
        "scan_info": "My-brother",
        "title": "The Smell of Burnt Toast",
        "original_format": "digital",
    },
    "The_Arkenstone_v03_(2002)_(Digital)_(DR_&amp;_Quenya-Elves).cbr": {
        "series": "The Arkenstone",
        "volume": "03",
        "issue": "03",
        "year": "2002",
        "ext": "cbr",
        "scan_info": "DR &amp; Quenya-Elves",
        "original_format": "Digital",
    },
    "Kartalk v01 001 - Fear the Brakes (2004) (digital) (Son of Ultron-EMpire).cbr": {
        "series": "Kartalk",
        "volume": "01",
        "issue": "001",
        "year": "2004",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-EMpire",
        "title": "Fear the Brakes",
    },
    "Kartalk Library Edition v01 (1992) (digital) (Son of Ultron-Empire).cbr": {
        "series": "Kartalk Library Edition",
        "volume": "01",
        "issue": "01",
        "year": "1992",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-Empire",
    },
    "Kind of Deadly v02 - Last Bullet (2006) (Digital) (Zone-Empire).cbr": {
        "series": "Kind of Deadly",
        "volume": "02",
        "issue": "02",
        "year": "2006",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
        "title": "Last Bullet",
    },
    "Jeremy John - Not A Title (2017) (digital-Minutement).cbz": {
        "series": "Jeremy John - Not A Title",
        "year": "2017",
        "ext": "cbz",
        "original_format": "digital",
        "scan_info": "Minutement",
    },
    "Jeremy John 001 (2006) (digital (Minutemen-Faessla).cbz": {
        "series": "Jeremy John",
        "issue": "001",
        "year": "2006",
        "ext": "cbz",
        "scan_info": "Minutemen-Faessla",
        "original_format": "digital",
    },
    "Jeremy John 003 (2007) (4 covers) (digital) (Minutemen-Faessla).cbz": {
        "series": "Jeremy John",
        "issue": "003",
        "year": "2007",
        "ext": "cbz",
        "scan_info": "Minutemen-Faessla",
        "original_format": "digital",
        "remainders": ("(4 covers)",),
    },
    "Jeremy John v01 - Uninterested! (2007) (Digital) (Asgard-Empire).cbr": {
        "series": "Jeremy John",
        "volume": "01",
        "issue": "01",
        "year": "2007",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Asgard-Empire",
        "title": "Uninterested!",
    },
    "King of Skittles 01 (of 05) (2020) (digital) (Son of Ultron-Empire).cbr": {
        "series": "King of Skittles",
        "issue": "01",
        "issue_count": "05",
        "year": "2020",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-Empire",
    },
    "Darkwad 011 (2019) (Digital) (Zone-Empire).cbr": {
        "series": "Darkwad",
        "issue": "011",
        "year": "2019",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
    },
    "Darkwad by Carlos Zemo v01 - Knuckle Fight (2009) (Digital) (Zone-Empire).cbr": {
        "series": "Darkwad by Carlos Zemo",
        "volume": "01",
        "issue": "01",
        "year": "2009",
        "ext": "cbr",
        "title": "Knuckle Fight",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
    },
    "The Walking Dead #002 (2003).cbz": {
        "series": "The Walking Dead",
        "issue": "002",
        "year": "2003",
        "ext": "cbz",
    },
    "The Walking Dead #3.cbz": {
        "series": "The Walking Dead",
        "issue": "3",
        "ext": "cbz",
    },
    "The Walking Dead 4.cbz": {
        "series": "The Walking Dead",
        "issue": "4",
        "ext": "cbz",
    },
    "A Fractional Comic 1.1.cbz": {
        "series": "A Fractional Comic",
        "issue": "1.1",
        "ext": "cbz",
    },
    "A Fractional Comic 8.54.cbz": {
        "series": "A Fractional Comic",
        "issue": "8.54",
        "ext": "cbz",
    },
    "Earth X #½.cbz": {
        "series": "Earth X",
        "issue": "½",
        "ext": "cbz",
    },
    "Avengers #001½.cbz": {
        "series": "Avengers",
        "issue": "001½",
        "ext": "cbz",
    },
    "The Amazing Spider-Man #78.BEY.cbz": {
        "series": "The Amazing Spider-Man",
        "issue": "78.BEY",
        "ext": "cbz",
    },
    "The Amazing Spider-Man #54.LR.cbz": {
        "series": "The Amazing Spider-Man",
        "issue": "54.LR",
        "ext": "cbz",
    },
    "Wolverine & the X-Men #27AU.cbz": {
        "series": "Wolverine & the X-Men",
        "issue": "27AU",
        "ext": "cbz",
    },
    "Fantastic Four #5AU.cbz": {
        "series": "Fantastic Four",
        "issue": "5AU",
        "ext": "cbz",
    },
}

# Tests for 0.2.0
FNS.update(
    {
        # Philosopy change regarding dashes.
        "Bardude - The Last Thing I Remember.cbz": {
            "series": "Bardude - The Last Thing I Remember",
            "ext": "cbz",
        },
        "Drunkguy - The Man Without Fear - 01.cbz": {
            "series": "Drunkguy - The Man Without Fear",
            "issue": "01",
            "ext": "cbz",
        },
        # BIG Change. title after token. more stripping.
        "'Batman - Superman - World's Finest 022 (2024) (Webrip) (The Last Kryptonian-DCP).cbz": {
            "ext": "cbz",
            "issue": "022",
            "original_format": "Webrip",
            "series": "Batman - Superman - World's Finest",
            "scan_info": "The Last Kryptonian-DCP",
            "year": "2024",
        },
        # Issue number starting with a letter requested in https://github.com/comictagger/comictagger/issues/543
        #   word characters now allowed to lead issue numbers only if preceded by a # marker
        "batman #B01 title.cbz": {
            "ext": "cbz",
            "issue": "B01",
            "series": "batman",
            "title": "title",
        },
        "Monster_Island_v1_#2__repaired__c2c.cbz": {
            "ext": "cbz",
            "issue": "2",
            "series": "Monster Island",
            "volume": "1",
            "scan_info": "c2c",
            "remainders": ("(repaired)",),
        },
        # Extra - in the series
        " X-Men-V1-#067.cbr": {
            "ext": "cbr",
            "issue": "067",
            "series": "X-Men",
            "volume": "1",
            "remainders": ("-",),
        },
        "Aquaman - Green Arrow - Deep Target #01 (of 07) (2021).cbr": {
            "ext": "cbr",
            "issue": "01",
            "series": "Aquaman - Green Arrow - Deep Target",
            "year": "2021",
            "issue_count": "07",
        },
        # CT only separates this into a title if the '-' is attached to the previous word eg 'aquaman- Green Arrow'. @bpepple opened a ticket for this https://github.com/ajslater/comicfn2dict/issues/1 already
        "Batman_-_Superman_#020_(2021).cbr": {
            "ext": "cbr",
            "issue": "020",
            "series": "Batman - Superman",
            "year": "2021",
        },
        # Publishers like to re-print some of their annuals using this format for the year
        "Batman '89 (2021) .cbr": {
            "ext": "cbr",
            "series": "Batman '89",
            "year": "2021",
        },
        # This made the parser in CT much more complicated. It's understandable that this isn't parsed on the first few iterations of this project
        "Star Wars - War of the Bounty Hunters - IG-88 (2021).cbz": {
            "ext": "cbz",
            "series": "Star Wars - War of the Bounty Hunters - IG-88",
            "year": "2021",
        },  # The addition of the '#1' turns this into the same as 'Aquaman - Green Arrow - Deep Target' above
        "Star Wars - War of the Bounty Hunters - IG-88 #1 (2021).cbz": {
            "ext": "cbz",
            "issue": "1",
            "series": "Star Wars - War of the Bounty Hunters - IG-88",
            "year": "2021",
        },
        "Free Comic Book Day - Avengers.Hulk (2021).cbz": {
            "ext": "cbz",
            "series": "Free Comic Book Day - Avengers Hulk",
            "year": "2021",
        },
        # CT assumes the volume is also the issue number if it can't find an issue number
        "Avengers By Brian Michael Bendis volume 03 (2013).cbz": {
            "ext": "cbz",
            "issue": "03",
            "series": "Avengers By Brian Michael Bendis",
            "volume": "03",
            "year": "2013",
        },
        # CT catches the year
        "Marvel Previews #002 (January 2022).cbr": {
            "ext": "cbr",
            "issue": "002",
            "series": "Marvel Previews",
            "publisher": "Marvel",
            "month": "01",
            "year": "2022",
        },
        "Test Numeric Year #2 2001-02-24.cbz": {
            "ext": "cbz",
            "issue": "2",
            "series": "Test Numeric Year",
            "year": "2001",
            "month": "02",
            "day": "24",
        },
        "Test Month First Date 02-24-2001.cbz": {
            "ext": "cbz",
            "series": "Test Month First Date",
            "year": "2001",
            "month": "02",
            "day": "24",
        },
        # CT notices that this is a full date, CT doesn't actually return the month or day though just removes it
        "X-Men, 2021-08-04 (#02).cbz": {
            "ext": "cbz",
            "issue": "02",
            "series": "X-Men",
            "year": "2021",
            "month": "08",
            "day": "04",
        },
        # 4 digit issue number
        #   should this be an issue number if year DONE?.
        "action comics 1024.cbz": {
            "ext": "cbz",
            "issue": "1024",
            "series": "action comics",
        },
        # This is a contrived test case. I've never seen this I just wanted to handle it with my parser
        "Cory Doctorow's Futuristic Tales of the Here and Now #0.0.1 (2007).cbz": {
            "ext": "cbz",
            "issue": "0.0.1",
            "series": "Cory Doctorow's Futuristic Tales of the Here and Now",
            "year": "2007",
        },
        # CT treats ':' the same as '-' but here the ':' is attached to 'Now' which CT sees as a title separation
        "Cory Doctorow's Futuristic Tales of the Here and Now: Anda's Game #001 (2007).cbz": {
            "ext": "cbz",
            "issue": "001",
            "series": "Cory Doctorow's Futuristic Tales of the Here and Now",
            "title": "Anda's Game",
            "year": "2007",
        },
        # If a title ends in a year, it's not an issue (and is a year if no year)
        "Blade Runner Free Comic Book Day 2021 (2021).cbr": {
            "ext": "cbr",
            "series": "Blade Runner Free Comic Book Day 2021",
            "year": "2021",
        },
        # If a year occurs after another year,  and no volume, do volume / year
        "Super Strange Yarns (1957) #92 (1969).cbz": {
            "ext": "cbz",
            "issue": "92",
            "series": "Super Strange Yarns",
            "volume": "1957",
            "year": "1969",
        },
        # CT checks for the following '(of 06)' after the '03' and marks it as the volume
        "Elephantmen 2259 #008 - Simple Truth 03 (of 06) (2021).cbr": {
            "ext": "cbr",
            "issue": "008",
            "series": "Elephantmen 2259",
            "title": "Simple Truth",
            "volume": "03",
            "year": "2021",
            "volume_count": "06",
        },
        # CT treats book like 'v' but also adds it as the title (matches ComicVine for this particular series)
        "Bloodshot Book 03 (2020).cbr": {
            "ext": "cbr",
            "issue": "03",
            "series": "Bloodshot",
            "title": "Book 03",
            "volume": "03",
            "year": "2020",
        },
        # c2c aka "cover to cover" is fairly common and CT moves it to scan_info/remainder
        "Marvel Two In One V1 #090  c2c.cbr": {
            "ext": "cbr",
            "issue": "090",
            "series": "Marvel Two In One",
            "publisher": "Marvel",
            "volume": "1",
            "scan_info": "c2c",
        },
        # CT treats '[]' as equivalent to '()', catches DC as a publisher and 'Sep-Oct 1951' as dates and removes them. CT doesn't catch the digital though so that could be better but I blame whoever made this atrocious filename
        "Wonder Woman #49 DC Sep-Oct 1951 digital [downsized, lightened, 4 missing story pages restored] (Shadowcat-Empire).cbz": {
            "ext": "cbz",
            "issue": "49",
            "series": "Wonder Woman",
            "publisher": "DC",
            "year": "1951",
            "month": "09",
            "remainders": (
                "digital (downsized, lightened, 4 missing story pages "
                "restored) (Shadowcat-Empire)",
            ),
        },
        "Captain Science #001 (1950) The Beginning - nothing.cbz": {
            "ext": "cbz",
            "issue": "001",
            "title": "The Beginning - nothing",
            "series": "Captain Science",
            "year": "1950",
        },
        "Captain Science #001-cix-cbi.cbr": {
            "ext": "cbr",
            "issue": "001",
            "series": "Captain Science",
            "title": "cix-cbi",
        },
        "Long Series Name v1 (2000) #001 "
        "Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS_VOL,
        "Long Series Name 001 (2000) (TPB-Releaser) Title.cbz": {
            "series": "Long Series Name",
            "issue": "001",
            "year": "2000",
            "original_format": "TPB",
            "scan_info": "Releaser",
            "remainders": ("Title",),
            "ext": "cbz",
        },
        "Long Series Name Vol 1 "
        "(2000) (TPB) (Releaser & Releaser-Releaser) Title.cbr": {
            "series": "Long Series Name",
            "volume": "1",
            "issue": "1",
            "remainders": ("Title",),
            "original_format": "TPB",
            "year": "2000",
            "scan_info": "Releaser & Releaser-Releaser",
            "ext": "cbr",
        },
    }
)

# first_key, first_val = NEW.popitem()
# FNS[first_key] = first_val
PARSE_FNS = MappingProxyType(FNS)

SERIALIZE_FNS = MappingProxyType(
    {
        "Long Series Name #001 (2000) Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
        "Long Series Name v1 #001 "
        "(2000) Title (TPB) (Releaser & Releaser-Releaser).cbr": TEST_COMIC_VOL_ONLY,
        "Series Name (2000-12-31).cbz": {
            "series": "Series Name",
            "year": "2000",
            "month": "12",
            "day": "31",
            "ext": "cbz",
        },
        "Series Name (2000-12).cbz": {
            "series": "Series Name",
            "year": "2000",
            "month": "12",
            "ext": "cbz",
        },
        "Series Name (Dec-31).cbz": {
            "series": "Series Name",
            "month": "12",
            "day": "31",
            "ext": "cbz",
        },
    }
)



"""Parse comic book archive names using the simple 'parse' parser."""

from __future__ import annotations

from calendar import month_abbr
from copy import copy
from pathlib import Path
from pprint import pformat
from sys import maxsize
from typing import TYPE_CHECKING

from comicfn2dict.log import print_log_header
from comicfn2dict.regex import (
    ALPHA_MONTH_RANGE_RE,
    BOOK_VOLUME_RE,
    ISSUE_BEGIN_RE,
    ISSUE_END_RE,
    ISSUE_NUMBER_RE,
    ISSUE_WITH_COUNT_RE,
    MONTH_FIRST_DATE_RE,
    NON_NUMBER_DOT_RE,
    ORIGINAL_FORMAT_SCAN_INFO_RE,
    ORIGINAL_FORMAT_SCAN_INFO_SEPARATE_RE,
    PUBLISHER_AMBIGUOUS_RE,
    PUBLISHER_AMBIGUOUS_TOKEN_RE,
    PUBLISHER_UNAMBIGUOUS_RE,
    PUBLISHER_UNAMBIGUOUS_TOKEN_RE,
    REGEX_SUBS,
    REMAINING_GROUP_RE,
    SCAN_INFO_SECONDARY_RE,
    TOKEN_DELIMETER,
    VOLUME_RE,
    VOLUME_WITH_COUNT_RE,
    YEAR_END_RE,
    YEAR_FIRST_DATE_RE,
    YEAR_TOKEN_RE,
)

if TYPE_CHECKING:
    from re import Match, Pattern

_DATE_KEYS = frozenset({"year", "month", "day"})
_REMAINING_GROUP_KEYS = ("series", "title")
# Ordered by commonness.
_TITLE_PRECEDING_KEYS = ("issue", "year", "volume", "month")


class ComicFilenameParser:
    """Parse a filename metadata into a dict."""

    def path_index(self, key: str, default: int = -1) -> int:
        """Lazily retrieve and memoize the key's location in the path."""
        if key == "remainders":
            return default
        value: str = self.metadata.get(key, "")  # type: ignore
        if not value:
            return default
        if value not in self._path_indexes:
            # XXX This is fragile, but it's difficult to calculate the original
            #     position at match time from the ever changing _unparsed_path.
            index = self.path.rfind(value) if key == "ext" else self.path.find(value)
            self._path_indexes[value] = index
        return self._path_indexes[value]

    def _log(self, label: str) -> None:
        if not self._debug:
            return
        print_log_header(label)
        combined = {}
        for key in self.metadata:
            combined[key] = (self.metadata.get(key), self.path_index(key))
        print("  " + self._unparsed_path)  # noqa: T201
        print("  " + pformat(combined))  # noqa: T201

    def _parse_ext(self) -> None:
        """Pop the extension from the pathname."""
        path = Path(self._unparsed_path)
        suffix = path.suffix
        if not suffix:
            return

        data = path.name.removesuffix(suffix)
        ext = suffix.lstrip(".")
        self.metadata["ext"] = ext
        self._unparsed_path = data

    def _clean_dividers(self) -> None:
        """Replace non space dividers and clean extra spaces out of string."""
        data = self._unparsed_path

        # Simple substitutions
        for regex, pair in REGEX_SUBS.items():
            replacement, count = pair
            data = regex.sub(replacement, data, count=count).strip()
        self._unparsed_path = data.strip()
        self._log("After Clean Path")

    def _parse_items_update_metadata(
        self, matches: Match, exclude: str, require_all: bool, first_only: bool
    ) -> bool:
        """Update Metadata."""
        matched_metadata = {}
        for key, value in matches.groupdict().items():
            if value == exclude:
                continue
            if not value:
                if require_all:
                    return False
                continue
            matched_metadata[key] = value
            if first_only:
                break
        if not matched_metadata:
            return False
        self.metadata.update(matched_metadata)
        return True

    def _parse_items_pop_tokens(self, regex: Pattern, first_only: bool) -> None:
        """Pop tokens from unparsed path."""
        count = 1 if first_only else 0
        marked_str = regex.sub(TOKEN_DELIMETER, self._unparsed_path, count=count)
        parts = []
        for part in marked_str.split(TOKEN_DELIMETER):
            if token := part.strip():
                parts.append(token)
        self._unparsed_path = TOKEN_DELIMETER.join(parts)

    def _parse_items(  # noqa: PLR0913
        self,
        regex: Pattern,
        require_all: bool = False,
        exclude: str = "",
        first_only: bool = False,
        pop: bool = True,
    ) -> None:
        """Parse a value from the data list into metadata and alter the data list."""
        # Match
        matches = regex.search(self._unparsed_path)
        if not matches:
            return

        if not self._parse_items_update_metadata(
            matches, exclude, require_all, first_only
        ):
            return

        if pop:
            self._parse_items_pop_tokens(regex, first_only)

    def _parse_issue(self) -> None:
        """Parse Issue."""
        self._parse_items(ISSUE_NUMBER_RE)
        if "issue" not in self.metadata:
            self._parse_items(ISSUE_WITH_COUNT_RE)
        self._log("After Issue")

    def _parse_volume(self) -> None:
        """Parse Volume."""
        self._parse_items(VOLUME_RE)
        if "volume" not in self.metadata:
            self._parse_items(VOLUME_WITH_COUNT_RE)
        self._log("After Volume")

    def _alpha_month_to_numeric(self) -> None:
        """Translate alpha_month to numeric month."""
        if alpha_month := self.metadata.pop("alpha_month", ""):
            alpha_month = alpha_month.capitalize()  # type: ignore
            for index, abbr in enumerate(month_abbr):
                if abbr and alpha_month.startswith(abbr):
                    month = f"{index:02d}"
                    self.metadata["month"] = month
                    break

    def _parse_dates(self) -> None:
        """Parse date schemes."""
        # Discard second month of alpha month ranges.
        self._unparsed_path = ALPHA_MONTH_RANGE_RE.sub(r"\1", self._unparsed_path)

        # Month first date
        self._parse_items(MONTH_FIRST_DATE_RE)
        self._alpha_month_to_numeric()

        # Year first date
        if _DATE_KEYS - self.metadata.keys():
            self._parse_items(YEAR_FIRST_DATE_RE)
            self._alpha_month_to_numeric()

        if "year" not in self.metadata:
            self._parse_items(YEAR_TOKEN_RE, first_only=True)
            if "volume" in self.metadata:
                return
            # A second year will be the real year.
            # Move the first year to volume
            if volume := self.metadata.get("year", ""):
                self._parse_items(YEAR_TOKEN_RE)
                if self.metadata.get("year", "") != volume:
                    self.metadata["volume"] = volume
        self._log("After Date")

    def _parse_format_and_scan_info(self) -> None:
        """Format & Scan Info."""
        self._parse_items(
            ORIGINAL_FORMAT_SCAN_INFO_RE,
            require_all=True,
        )
        if "original_format" not in self.metadata:
            self._parse_items(
                ORIGINAL_FORMAT_SCAN_INFO_SEPARATE_RE,
            )
        self._parse_items(SCAN_INFO_SECONDARY_RE)
        if (
            scan_info_secondary := self.metadata.pop("secondary_scan_info", "")
        ) and "scan_info" not in self.metadata:
            self.metadata["scan_info"] = scan_info_secondary  # type: ignore
        self._log("After original_format & scan_info")

    def _parse_ends_of_remaining_tokens(self):
        # Volume left on the end of string tokens
        if "volume" not in self.metadata:
            self._parse_items(BOOK_VOLUME_RE)
            self._log("After original_format & scan_info")

        # Years left on the end of string tokens
        year_end_matched = False
        if "year" not in self.metadata:
            self._parse_items(YEAR_END_RE, pop=False)
            year_end_matched = "year" in self.metadata
            self._log("After Year on end of token")

        # Issue left on the end of string tokens
        if "issue" not in self.metadata and not year_end_matched:
            exclude: str = self.metadata.get("year", "")  # type: ignore
            self._parse_items(ISSUE_END_RE, exclude=exclude)
        if "issue" not in self.metadata:
            self._parse_items(ISSUE_BEGIN_RE)
        self._log("After Issue on ends of tokens")

    def _parse_publisher(self) -> None:
        """Parse Publisher."""
        # Pop single tokens so they don't end up titles.
        self._parse_items(PUBLISHER_UNAMBIGUOUS_TOKEN_RE, first_only=True)
        if "publisher" not in self.metadata:
            self._parse_items(PUBLISHER_AMBIGUOUS_TOKEN_RE, first_only=True)
        if "publisher" not in self.metadata:
            self._parse_items(PUBLISHER_UNAMBIGUOUS_RE, pop=False, first_only=True)
        if "publisher" not in self.metadata:
            self._parse_items(PUBLISHER_AMBIGUOUS_RE, pop=False, first_only=True)
        self._log("After publisher")

    def _is_at_title_position(self, value: str) -> bool:
        """Title is in correct position."""
        title_index = self.path.find(value)

        # Titles must come after series but before format and scan_info
        if (
            title_index < self.path_index("series")
            or title_index > self.path_index("original_format", maxsize)
            or title_index > self.path_index("scan_info", maxsize)
        ):
            return False

        # Titles must be after the series and one other token.
        title_ok = False
        other_tokens_exist = False
        for preceding_key in _TITLE_PRECEDING_KEYS:
            other_tokens_exist = True
            if title_index > self.path_index(preceding_key):
                title_ok = True
                break
        return title_ok or not other_tokens_exist

    def _grouping_operators_strip(self, value: str) -> str:
        """Strip spaces and parens."""
        value = value.strip()
        value = value.strip("()").strip()
        value = value.strip("-").strip()
        value = value.strip(",").strip()
        value = value.strip("'").strip()
        return value.strip('"').strip()

    def _parse_series_and_title_token(
        self, remaining_key_index: int, tokens: list[str]
    ) -> str:
        """Parse one series or title token."""
        key = _REMAINING_GROUP_KEYS[remaining_key_index]
        if key in self.metadata:
            return ""
        token = tokens.pop(0)
        match = REMAINING_GROUP_RE.search(token)
        if not match:
            return token
        value = match.group()
        if key == "title" and not self._is_at_title_position(value):
            return token
        value = NON_NUMBER_DOT_RE.sub(r"\1 \2", value)
        value = self._grouping_operators_strip(value)
        if value:
            self.metadata[key] = value
        return ""

    def _parse_series_and_title(self) -> None:
        """Assign series and title."""
        if not self._unparsed_path:
            return

        remaining_key_index = 0
        unused_tokens = []
        tokens = self._unparsed_path.split(TOKEN_DELIMETER)
        while tokens and remaining_key_index < len(_REMAINING_GROUP_KEYS):
            unused_token = self._parse_series_and_title_token(
                remaining_key_index, tokens
            )
            if unused_token:
                unused_tokens.append(unused_token)
            remaining_key_index += 1

        self._unparsed_path = " ".join(unused_tokens) if unused_tokens else ""
        self._log("After Series & Title")

    def _add_remainders(self) -> None:
        """Add Remainders."""
        remainders = []
        for token in self._unparsed_path.split(TOKEN_DELIMETER):
            if remainder := token.strip():
                remainders.append(remainder)

        if remainders:
            self.metadata["remainders"] = tuple(remainders)

    def parse(self) -> dict[str, str | tuple[str, ...]]:
        """Parse the filename with a hierarchy of regexes."""
        self._log("Init")
        self._parse_ext()
        self._clean_dividers()
        self._parse_issue()
        self._parse_volume()
        self._parse_dates()
        self._parse_format_and_scan_info()
        self._parse_ends_of_remaining_tokens()
        self._parse_publisher()
        self._parse_series_and_title()

        # Copy volume into issue if it's all we have.
        if "issue" not in self.metadata and "volume" in self.metadata:
            self.metadata["issue"] = self.metadata["volume"]
        self._log("After issue can be volume")

        self._add_remainders()

        return self.metadata

    def __init__(self, path: str | Path, verbose: int = 0):
        """Initialize."""
        self._debug: bool = verbose > 0
        # munge path
        if isinstance(path, str):
            path = path.strip()
        p_path = Path(path)
        self.path = str(p_path.name).strip()
        self.metadata: dict[str, str | tuple[str, ...]] = {}
        self._unparsed_path = copy(self.path)
        self._path_indexes: dict[str, int] = {}


def comicfn2dict(
    path: str | Path, verbose: int = 0
) -> dict[str, str | tuple[str, ...]]:
    """Simplfily the API."""
    parser = ComicFilenameParser(path, verbose=verbose)
    return parser.parse()



"""Unparse comic filenames."""

from __future__ import annotations

from calendar import month_abbr
from collections.abc import Callable, Mapping, Sequence
from contextlib import suppress
from types import MappingProxyType

from comicfn2dict.log import print_log_header


def issue_formatter(issue: str) -> str:
    """Formatter to zero pad issues."""
    i = 0
    issue = issue.lstrip("0")
    for c in issue:
        if not c.isdigit():
            break
        i += 1
    pad = 3 + len(issue) - i
    return "#{:0>" + str(pad) + "}"


_PAREN_FMT: str = "({})"
_FILENAME_FORMAT_TAGS: tuple[tuple[str, str | Callable], ...] = (
    ("series", "{}"),
    ("volume", "v{}"),
    ("volume_count", "(of {:03})"),
    ("issue", issue_formatter),
    ("issue_count", "(of {:03})"),
    ("date", _PAREN_FMT),
    ("title", "{}"),
    ("publisher", _PAREN_FMT),
    ("original_format", _PAREN_FMT),
    ("scan_info", _PAREN_FMT),
)
_EMPTY_VALUES: tuple[None, str] = (None, "")
_DEFAULT_EXT = "cbz"
_DATE_KEYS = ("year", "month", "day")


class ComicFilenameSerializer:
    """Serialize Comic Filenames from dict."""

    def _log(self, label: str, fn: str) -> None:
        """Log progress."""
        if not self._debug:
            return
        print_log_header(label)
        print(fn)  # noqa: T201

    def _add_date(self) -> None:
        """Construct date from Y-m-D if they exist."""
        if "date" in self.metadata:
            return
        parts = []
        for key in _DATE_KEYS:
            if part := self.metadata.get(key):
                if key == "month" and not parts:
                    with suppress(TypeError):
                        part = month_abbr[int(part)]

                parts.append(part)
            if key == "month" and not parts:
                # noop if only day.
                break
        if parts:
            parts = (str(part) for part in parts)
            date = "-".join(parts)
            self._log("After date", date)
            self.metadata = MappingProxyType({**self.metadata, "date": date})

    def _tokenize_tag(self, tag: str, fmt: str | Callable) -> str:
        """Add tags to the string."""
        val = self.metadata.get(tag)
        if val in _EMPTY_VALUES:
            return ""
        final_fmt = fmt(val) if isinstance(fmt, Callable) else fmt
        return final_fmt.format(val).strip()

    def _add_remainder(self) -> str:
        """Add the remainders specially."""
        if remainders := self.metadata.get("remainders"):
            if isinstance(remainders, Sequence):
                remainders = (str(remainder) for remainder in remainders)
                remainder = " ".join(remainders)
            else:
                remainder = str(remainders)
            return f"[{remainder}]"
        return ""

    def serialize(self) -> str:
        """Get our preferred basename from a metadata dict."""
        self._add_date()

        tokens = []
        for tag, fmt in _FILENAME_FORMAT_TAGS:
            if token := self._tokenize_tag(tag, fmt):
                tokens.append(token)
            self._log(f"After {tag}", str(tokens))
        fn = " ".join(tokens)

        fn += self._add_remainder()
        self._log("After remainder", fn)

        if self._ext:
            ext = self.metadata.get("ext", _DEFAULT_EXT)
            fn += f".{ext}"
            self._log("After ext", fn)

        return fn

    def __init__(self, metadata: Mapping, ext: bool = True, verbose: int = 0):
        """Initialize."""
        self.metadata: Mapping = metadata
        self._ext: bool = ext
        self._debug: bool = bool(verbose)


def dict2comicfn(md: Mapping, ext: bool = True, verbose: int = 0) -> str:
    """Simplify API."""
    serializer = ComicFilenameSerializer(md, ext=ext, verbose=verbose)
    return serializer.serialize()



"""Tests for filename parsing."""

from pprint import pprint

import pytest
from deepdiff.diff import DeepDiff

from comicfn2dict import ComicFilenameParser
from tests.comic_filenames import PARSE_FNS


@pytest.mark.parametrize("item", PARSE_FNS.items())
def test_parse_filename(item):
    """Test filename parsing."""
    fn, defined_fields = item
    md = ComicFilenameParser(fn, verbose=1).parse()
    diff = DeepDiff(defined_fields, md, ignore_order=True)
    print(fn)
    pprint(defined_fields)
    pprint(md)
    pprint(diff)
    assert not diff



"""Print log header."""


def print_log_header(label: str) -> None:
    """Print log header."""
    prefix = "-" * 3 + label
    suffix_len = 80 - len(prefix)
    suffix = "-" * suffix_len
    print(prefix + suffix)  # noqa: T201



#!/usr/bin/env python3
"""Simple cli for comicfn2dict."""

from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint

from comicfn2dict.parse import ComicFilenameParser


def main() -> None:
    """Test parser."""
    description = "Comic book archive read/write tool."
    parser = ArgumentParser(description=description)
    parser.add_argument("path", help="Path of comic filename to parse", type=Path)
    parser.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        help="Display intermediate parsing steps. Good for debugging.",
    )
    args = parser.parse_args()
    name = args.path.name
    cfnparser = ComicFilenameParser(name, verbose=args.verbose)
    metadata = cfnparser.parse()
    if args.verbose:
        print("=" * 80)  # noqa:T201
    pprint(metadata)  # noqa:T203


if __name__ == "__main__":
    main()



"""Parsing regexes."""

from re import IGNORECASE, Pattern, compile
from types import MappingProxyType

PUBLISHERS_UNAMBIGUOUS: tuple[str, ...] = (
    r"Abrams ComicArts",
    r"BOOM! Studios",
    r"DC(\sComics)?",
    r"Dark Horse Comics",
    r"Drawn & Quarterly",
    r"Dynamite Entertainment",
    r"IDW Publishing",
    r"Icon Comics",
    r"Kodansha",
    r"Oni Press",
    r"Pantheon Books",
    r"SLG Publishing",
    r"SelfMadeHero",
    r"Titan Comics",
)
PUBLISHERS_AMBIGUOUS: tuple[str, ...] = (
    r"(?<!Capt\.\s)(?<!Capt\s)(?<!Captain\s)Marvel",
    r"Heavy Metal",
    r"Epic",
    r"Image",
    r"Mirage",
)

ORIGINAL_FORMAT_PATTERNS: tuple[str, ...] = (
    r"Anthology",
    r"(One|1)[-\s]Shot",
    r"Annual",
    r"Annotation[s]?",
    r"Box[-\s]Set",
    r"Digital",
    r"Director[’']?s\sCut",  # noqa: RUF001
    r"Giant([-\s]Size(d)?)?",
    r"Graphic\sNovel",
    r"GN",
    r"Hard[-\s]?Cover",
    r"HC",
    r"HD-Upscaled",
    r"King[-\s]Size(d)?",
    r"Magazine",
    r"Manga?",
    r"Omnibus",
    r"PDF([-\s]Rip)?",
    r"Preview",
    r"Prologue",
    r"Scanlation",
    r"Script",
    r"Sketch",
    r"TPB",
    r"Trade[-\s]Paper[-\s]?Back",
    r"Web([-\s]?(Comic|Rip))?",
)

MONTHS: tuple[str, ...] = (
    r"Jan(uary)?",
    r"Feb(ruary)?",
    r"Mar(ch)?",
    r"Apr(il)?",
    r"May",
    r"Jun(e)?",
    r"Jul(y)?",
    r"Aug(ust)?",
    r"Sep(tember)?",
    r"Oct(ober)?",
    r"Nov(ember)?",
    r"Dec(ember)?",
)

TOKEN_DELIMETER: str = r"/"


def re_compile(exp: str, parenthify: bool = False) -> Pattern:
    """Compile regex with options."""
    if parenthify:
        exp = r"\(" + exp + r"\)"
    return compile(exp, flags=IGNORECASE)


# CLEAN
_TOKEN_DIVIDERS_RE = re_compile(r":")
_SPACE_EQUIVALENT_RE = re_compile(r"_")
_EXTRA_SPACES_RE = re_compile(r"\s\s+")
_LEFT_PAREN_EQUIVALENT_RE = re_compile(r"\[")
_RIGHT_PAREN_EQUIVALENT_RE = re_compile(r"\]")
_DOUBLE_UNDERSCORE_RE = re_compile(r"__(.*)__")
REGEX_SUBS: MappingProxyType[Pattern, tuple[str, int]] = MappingProxyType(
    {
        _DOUBLE_UNDERSCORE_RE: (r"(\1)", 0),
        _TOKEN_DIVIDERS_RE: (TOKEN_DELIMETER, 1),
        _SPACE_EQUIVALENT_RE: (r" ", 0),
        _EXTRA_SPACES_RE: (r" ", 0),
        _LEFT_PAREN_EQUIVALENT_RE: (r"(", 0),
        _RIGHT_PAREN_EQUIVALENT_RE: (r")", 0),
    }
)

### DATES
_YEAR_RE_EXP = r"(?P<year>[12]\d{3})"
_MONTH_ALPHA_RE_EXP = r"(" + "(?P<alpha_month>" + r"|".join(MONTHS) + r")\.?" r")"
_MONTH_NUMERIC_RE_EXP = r"(?P<month>0?\d|1[0-2]?)"
_MONTH_RE_EXP = r"(" + _MONTH_ALPHA_RE_EXP + r"|" + _MONTH_NUMERIC_RE_EXP + r")"
_ALPHA_MONTH_RANGE = (
    r"\b"  # noqa: ISC003
    + r"("
    + r"|".join(MONTHS)
    + r")"
    + r"("
    + r"\.?-"
    + r"("
    + r"|".join(MONTHS)
    + r")"
    + r")\b"
)
ALPHA_MONTH_RANGE_RE: Pattern = re_compile(_ALPHA_MONTH_RANGE)

_DAY_RE_EXP = r"(?P<day>([0-2]?\d|(3)[0-1]))"
_DATE_DELIM = r"[-\s]+"
_MONTH_FIRST_DATE_RE_EXP = (
    r"((\b|\(?)"
    # Month
    + _MONTH_RE_EXP
    # Day
    + r"("
    + _DATE_DELIM
    + _DAY_RE_EXP
    + r")?"
    # Year
    + r"[,]?"
    + _DATE_DELIM
    + _YEAR_RE_EXP
    + r"(\)?|\b))"
)
_YEAR_FIRST_DATE_RE_EXP = (
    r"(\b\(?"
    + _YEAR_RE_EXP
    + _DATE_DELIM
    + _MONTH_RE_EXP
    + _DATE_DELIM
    + _DAY_RE_EXP
    + r"\b\)?)"
)

MONTH_FIRST_DATE_RE: Pattern = re_compile(_MONTH_FIRST_DATE_RE_EXP)
YEAR_FIRST_DATE_RE: Pattern = re_compile(_YEAR_FIRST_DATE_RE_EXP)
YEAR_TOKEN_RE: Pattern = re_compile(_YEAR_RE_EXP, parenthify=True)
YEAR_END_RE: Pattern = re_compile(_YEAR_RE_EXP + r"\/|$")

# PAREN GROUPS
_OF_PATTERNS = r"|".join(ORIGINAL_FORMAT_PATTERNS)
_ORIGINAL_FORMAT_RE_EXP = r"(?P<original_format>" + _OF_PATTERNS + r")"
_SCAN_INFO_RE_EXP = r"(?P<scan_info>[^()]*)"
_ORIGINAL_FORMAT_SCAN_INFO_RE_EXP = (
    _ORIGINAL_FORMAT_RE_EXP + r"\s*[\(:-]" + _SCAN_INFO_RE_EXP  # + r")?"
)
# Keep this even though comicfn2dict doesn't use it directly
ORIGINAL_FORMAT_RE: Pattern = re_compile(_ORIGINAL_FORMAT_RE_EXP, parenthify=True)
ORIGINAL_FORMAT_SCAN_INFO_RE: Pattern = re_compile(
    _ORIGINAL_FORMAT_SCAN_INFO_RE_EXP, parenthify=True
)
ORIGINAL_FORMAT_SCAN_INFO_SEPARATE_RE: Pattern = re_compile(
    r"\(" + _ORIGINAL_FORMAT_RE_EXP + r"\).*\(" + _SCAN_INFO_RE_EXP + r"\)"
)

SCAN_INFO_SECONDARY_RE: Pattern = re_compile(r"\b(?P<secondary_scan_info>c2c)\b")

# ISSUE
_ISSUE_RE_EXP = r"(?P<issue>\w*(½|\d+)[\.\d+]*\w*)"
_ISSUE_COUNT_RE_EXP = r"\(of\s*(?P<issue_count>\d+)\)"
ISSUE_NUMBER_RE: Pattern = re_compile(
    r"(\(?#" + _ISSUE_RE_EXP + r"\)?)" + r"(\W*" + _ISSUE_COUNT_RE_EXP + r")?"
)
ISSUE_WITH_COUNT_RE: Pattern = re_compile(
    r"(\(?" + _ISSUE_RE_EXP + r"\)?" + r"\W*" + _ISSUE_COUNT_RE_EXP + r")"
)
ISSUE_END_RE: Pattern = re_compile(r"([\/\s]\(?" + _ISSUE_RE_EXP + r"\)?(\/|$))")
ISSUE_BEGIN_RE: Pattern = re_compile(r"((^|\/)\(?" + _ISSUE_RE_EXP + r"\)?[\/|\s])")

# Volume
_VOLUME_COUNT_RE_EXP = r"\(of\s*(?P<volume_count>\d+)\)"
VOLUME_RE: Pattern = re_compile(
    r"(" + r"(?:v(?:ol(?:ume)?)?\.?)\s*(?P<volume>\d+)"  # noqa: ISC003
    r"(\W*" + _VOLUME_COUNT_RE_EXP + r")?" + r")"
)
VOLUME_WITH_COUNT_RE: Pattern = re_compile(
    r"(\(?" + r"(?P<volume>\d+)" + r"\)?" + r"\W*" + _VOLUME_COUNT_RE_EXP + r")"
)
BOOK_VOLUME_RE: Pattern = re_compile(r"(?P<title>" + r"book\s*(?P<volume>\d+)" + r")")

# Publisher
_PUBLISHER_UNAMBIGUOUS_RE_EXP = (
    r"(\b(?P<publisher>" + r"|".join(PUBLISHERS_UNAMBIGUOUS) + r")\b)"
)
_PUBLISHER_AMBIGUOUS_RE_EXP = (
    r"(\b(?P<publisher>" + r"|".join(PUBLISHERS_AMBIGUOUS) + r")\b)"
)
PUBLISHER_UNAMBIGUOUS_TOKEN_RE: Pattern = re_compile(
    r"(^|\/)" + _PUBLISHER_UNAMBIGUOUS_RE_EXP + r"($|\/)"
)
PUBLISHER_AMBIGUOUS_TOKEN_RE: Pattern = re_compile(
    r"(^|\/)" + _PUBLISHER_AMBIGUOUS_RE_EXP + r"($|\/)"
)
PUBLISHER_UNAMBIGUOUS_RE: Pattern = re_compile(_PUBLISHER_UNAMBIGUOUS_RE_EXP)
PUBLISHER_AMBIGUOUS_RE = re_compile(_PUBLISHER_AMBIGUOUS_RE_EXP)

# LONG STRINGS
REMAINING_GROUP_RE: Pattern = re_compile(r"^[^\(].*[^\)]")
NON_NUMBER_DOT_RE: Pattern = re_compile(r"(\D)\.(\D)")
